
/*
 * Copyright (c) 1993 Daniel Boulet
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 *
 * Redistribution in binary form may occur without any restrictions.
 * Obviously, it would be nice if you gave credit where credit is due
 * but requiring it would be too onerous.
 *
 * This software is provided ``AS IS'' without any warranties of any kind.
 *
 */

/*
 *  Linux port (c) 1994 Bob Beck
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 *
 * This software is provided ``AS IS'' without any warranties of any kind.
 *
 */

/*
 * Implement IP packet firewall
 */


/* $Header: /usr/brule/dept/beck/ipfirewall/RCS/ip_firewall.c,v 1.2 1994/08/02 01:40:29 beck Exp $ */

/* $Log: ip_firewall.c,v $
 * Revision 1.2  1994/08/02  01:40:29  beck
 *  changed so LINUX not defined by default in this file so it won't break
 * the BSDI stuff. file added to linux kernel by the patch in the linux
 * dir will have it defined.
 *
 * Revision 1.1  1994/07/28  00:04:17  beck
 * Initial revision
 * */

/* #define LINUX Uncomment for Linux kernel, leave undefined otherwise */

#ifdef LINUX
#include <asm/segment.h>
#include <asm/system.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/socket.h>
#include <linux/sockios.h>
#include <linux/in.h>
#include <linux/inet.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include "ip.h"
#include "protocol.h"
#include "route.h"
#include "tcp.h"
#include <linux/skbuff.h>
#include "sock.h"
#include "arp.h"
#include "icmp.h"
#include <linux/ip_firewall.h>
#else
#include "param.h"
#include "systm.h"
#include "malloc.h"
#include "mbuf.h"
#include "domain.h"
#include "protosw.h"
#include "socket.h"
#include "errno.h"
#include "time.h"
#include "kernel.h"

#include "../net/if.h"
#include "../net/route.h"

#include "in.h"
#include "in_systm.h"
#include "ip.h"
#include "in_pcb.h"
#include "in_var.h"
#include "ip_var.h"
#include "ip_icmp.h"

#include "ip_firewall.h"
#endif

#undef DEBUG_IPFIREWALL

#ifdef LINUX
#define printf printk
#endif


struct ip_firewall *ip_firewall_forwarding_chain;
struct ip_firewall *ip_firewall_blocking_chain;

static struct ip_firewall **forwarding_chain_next = &ip_firewall_forwarding_chain;
static struct ip_firewall **blocking_chain_next = &ip_firewall_blocking_chain;

/*
 * Returns 1 if the port is matched by the vector, 0 otherwise
 */

#ifdef LINUX
/* Smegging Linux doesn't like K&R style prototypes in the kernel
   files,  since it compiles with gcc -Wall. Thought about
   doing this with #ifdef __GNUC__, but decided against it as 
   some functions in here use different args under linux than
   the BSD/386 version. Perhaps I can convince Danny to ANSIize 
   his code? :-) 
   -BB */
static int port_match(u_short *portptr, int nports, u_short port, int range_flag)
#else
static
port_match(portptr,nports,port,range_flag)
u_short *portptr;
int nports;
u_short port;
int range_flag;
#endif
{
    if ( range_flag ) {
	if ( portptr[0] <= port && port <= portptr[1] ) {
	    return( 1 );
	}
	nports -= 2;
	portptr += 2;
    }
    while ( nports-- > 0 ) {
	if ( *portptr++ == port ) {
	    return( 1 );
	}
    }
    return(0);

}

#ifdef LINUX
static void print_ip(unsigned long xaddr)
#else
static
void
print_ip(xaddr)
struct in_addr xaddr;
#endif
{
#ifdef LINUX
    unsigned long addr = ntohl(xaddr);
#else
    u_long addr = ntohl(xaddr.s_addr);
#endif
    printf("%d.%d.%d.%d",
	   (int)(addr>>24)&0xff,
	   (int)(addr>>16)&0xff,
	   (int)(addr>>8)&0xff,
	   (int) addr&0xff);
/*    printf ("%x",addr); */
}


#ifdef LINUX
int xip_firewall_check(struct iphdr *ip, struct ip_firewall *chain)
#else
xip_firewall_check(ip,chain)
struct ip *ip;
struct ip_firewall *chain;
#endif
{
#ifndef LINUX
    struct in_addr src, dst;
#endif
    char got_proto = 0;
    int firewall_proto, proto;
    register struct ip_firewall *fptr;
    u_short src_port, dst_port;

    if ( chain == NULL ) {	/* Is there a firewall chain? */
	return(1);
    }

#ifndef LINUX
    src = ip->ip_src;
    dst = ip->ip_dst;
#endif


#ifdef DEBUG_IPFIREWALL
    {
#ifdef LINUX
	u_short *portptr = (u_short *)&(((u_int *)ip)[ip->ihl]);
	printf("packet ");
	switch(ip->protocol) {
#else
	u_short *portptr = (u_short *)&(((u_int *)ip)[ip->ip_hl]);
	printf("packet ");
	switch(ip->ip_p) {
#endif
	case IPPROTO_TCP: printf("TCP "); break;
	case IPPROTO_UDP: printf("UDP "); break;
	case IPPROTO_ICMP: printf("ICMP:%d ",((char *)portptr)[0]&0xff); break;
#ifdef LINUX
	default: printf("p=%d ",ip->protocol); break;
	}
	print_ip(ip->saddr);
	if ( ip->protocol == IPPROTO_TCP || ip->protocol == IPPROTO_UDP ) {
#else
	default: printf("p=%d ",ip->ip_p); break;
	}
	print_ip(ip->ip_src);
	if ( ip->ip_p == IPPROTO_TCP || ip->ip_p == IPPROTO_UDP ) {
#endif
	    printf(":%d ",ntohs(portptr[0]));
	}
#ifdef LINUX
	print_ip(ip->daddr);
	if ( ip->protocol == IPPROTO_TCP || ip->protocol == IPPROTO_UDP ) {
#else
	print_ip(ip->ip_dst);
	if ( ip->ip_p == IPPROTO_TCP || ip->ip_p == IPPROTO_UDP ) {
#endif
	    printf(":%d ",ntohs(portptr[1]));
	}
	printf("\n");
    }
#endif

    for ( fptr = chain; fptr != NULL; fptr = fptr->next ) {

#ifdef LINUX
        /*  magic number check.. make sure what we have purports to be a
         *  struct ip_firewall and we didn't do a stupid or have someone
         *  clobber us from elsewhereville. Hence our kernel will
         *  panic if it finds that our structure has been corrupted
         *  -BB */
        if (fptr->magic != IP_FIREWALL_MAGIC) {
	  printf("ip_firewall_ctl: Bad magic number in ip_firewall_ctl, got %d, expecting %d\n",
		 fptr->magic,
		 IP_FIREWALL_MAGIC);
	  panic("ip_firewall_ctl: Bad magic number in ip_firewall structure!\n");
        }
      
	if ( (ip->saddr & fptr->src_mask) == fptr->src
	&&   (ip->daddr & fptr->dst_mask) == fptr->dst ) {
#else
	if ( (src.s_addr & fptr->src_mask.s_addr) == fptr->src.s_addr
	&&   (dst.s_addr & fptr->dst_mask.s_addr) == fptr->dst.s_addr ) {
#endif

	    if ( (firewall_proto = fptr->flags & IP_FIREWALL_KIND) == IP_FIREWALL_UNIVERSAL ) {

		/* Universal firewall - we've got a match! */

#ifdef DEBUG_IPFIREWALL
		printf("universal firewall match\n");
#endif
		return( (fptr->flags & IP_FIREWALL_ACCEPT) == IP_FIREWALL_ACCEPT );

	    } else {

		/* Specific firewall - packet's protocol must match firewall's */

		if ( !got_proto ) {
#ifdef LINUX
		    u_short *portptr = (u_short *)&(((u_int *)ip)[ip->ihl]);
		    switch( ip->protocol ) {
#else
		    u_short *portptr = (u_short *)&(((u_int *)ip)[ip->ip_hl]);
		    switch( ip->ip_p ) {
#endif
		    case IPPROTO_TCP:
			proto = IP_FIREWALL_TCP;
			src_port = ntohs(portptr[0]);	/* first two shorts in TCP */
			dst_port = ntohs(portptr[1]);	/* are src and dst ports */
			break;
		    case IPPROTO_UDP:
			proto = IP_FIREWALL_UDP;
			src_port = ntohs(portptr[0]);	/* first two shorts in UDP */
			dst_port = ntohs(portptr[1]);	/* are src and dst ports */
			break;
		    case IPPROTO_ICMP:
			proto = IP_FIREWALL_ICMP;
			break;
		    default: proto = IP_FIREWALL_UNIVERSAL;
#ifdef DEBUG_IPFIREWALL
			printf("non TCP/UDP packet\n");
#endif
		    }
		    got_proto = 1;
		}

		if ( proto == firewall_proto ) {

		    if (
			proto == IP_FIREWALL_ICMP
		    ||
			(
			    (
				fptr->num_src_ports == 0
			    ||
				port_match( &fptr->ports[0],
					    fptr->num_src_ports,
					    src_port,
					    fptr->flags & IP_FIREWALL_SRC_RANGE
					  )
			    )
			&&
			    (
				fptr->num_dst_ports == 0
			    ||
				port_match( &fptr->ports[fptr->num_src_ports],
					    fptr->num_dst_ports,
					    dst_port,
					    fptr->flags & IP_FIREWALL_DST_RANGE
					  )
			    )
			)
		    ) {

			return( (fptr->flags & IP_FIREWALL_ACCEPT) == IP_FIREWALL_ACCEPT );

		    }

		}

	    }

	}

     }

    /*
     * If we get here then none of the firewalls matched.
     * If the first firewall was an accept firewall then reject the packet.
     * If the first firewall was a deny firewall then accept the packet.
     *
     * The basic idea is that there is a virtual final firewall which is
     * the exact complement of the first firewall (this idea is a slight
     * variant of the way that the Telebit's Netblazer IP filtering scheme
     * handles this case).
     */

    return( ((chain->flags) & IP_FIREWALL_ACCEPT) != IP_FIREWALL_ACCEPT );

}


/*
 * Returns 0 if packet should be dropped, 1 if it should be accepted
 */
#ifdef LINUX
int ip_firewall_check(struct iphdr *ip, struct ip_firewall *chain)
#else
ip_firewall_check(ip,chain)
struct ip *ip;
struct ip_firewall *chain;
#endif
{
    if ( !xip_firewall_check(ip,chain) ) {
#ifdef LINUX
	u_short *portptr = (u_short *)&(((u_int *)ip)[ip->ihl]);
	printf("ip_firewall_check says no to ");
	switch(ip->protocol) {
#else
	u_short *portptr = (u_short *)&(((u_int *)ip)[ip->ip_hl]);
	printf("ip_firewall_check says no to ");
	switch(ip->ip_p) {
#endif
	case IPPROTO_TCP: printf("TCP "); break;
	case IPPROTO_UDP: printf("UDP "); break;
	case IPPROTO_ICMP: printf("ICMP:%d ",((char *)portptr)[0]&0xff); break;
#ifdef LINUX
	default: printk("p=%d ",ip->protocol); break;
	}
	print_ip(ip->saddr);
	if ( ip->protocol == IPPROTO_TCP || ip->protocol == IPPROTO_UDP ) {
#else
	default: printf("p=%d ",ip->ip_p); break;
	}
	print_ip(ip->ip_src);
	if ( ip->ip_p == IPPROTO_TCP || ip->ip_p == IPPROTO_UDP ) {
#endif
	    printf(":%d ",ntohs(portptr[0]));
	} else {
	    printf("\n");
	}
#ifdef LINUX
	print_ip(ip->daddr);
	if ( ip->protocol == IPPROTO_TCP || ip->protocol == IPPROTO_UDP ) {
#else
	print_ip(ip->ip_dst);
	if ( ip->ip_p == IPPROTO_TCP || ip->ip_p == IPPROTO_UDP ) {
#endif
	    printf(":%d ",ntohs(portptr[1]));
	}
	printf("\n");
	return(0);
    }
    return(1);
}


#ifdef LINUX
static void show_firewall_chain(char *name, struct ip_firewall *chain)
#else
static
void
show_firewall_chain(name,chain)
char *name;
struct ip_firewall *chain;
#endif
{
    if ( chain == NULL ) {
	printf("%s is empty\n",name);
    } else {
	printf("%s:\n",name);
	while ( chain != NULL ) {
	    char *comma;
	    int i;

	    printf("%x:  ",(int)chain);

	    if ( chain->flags & IP_FIREWALL_ACCEPT ) {
		printf("accept ");
	    } else {
		printf("deny   ");
	    }

	    switch ( chain->flags & IP_FIREWALL_KIND ) {
	    case IP_FIREWALL_ICMP: printf("icmp "); break;
	    case IP_FIREWALL_TCP: printf("tcp  "); break;
	    case IP_FIREWALL_UDP: printf("udp  "); break;
	    case IP_FIREWALL_UNIVERSAL: printf("all  "); break;
	    default: break;
	    }

	    print_ip(chain->src); printf(":");
	    print_ip(chain->src_mask);

	    comma = " ";
	    for ( i = 0; i < chain->num_src_ports; i += 1 ) {
		printf("%s%d",comma,chain->ports[i]);
		if ( i == 0 && (chain->flags & IP_FIREWALL_SRC_RANGE) ) {
		    comma = ":";
		} else {
		    comma = ",";
		}
	    }

	    printf(" ");
	    print_ip(chain->dst); printf(":");
	    print_ip(chain->dst_mask);

	    comma = " ";
	    for ( i = 0; i < chain->num_dst_ports; i += 1 ) {
		printf("%s%d",comma,chain->ports[chain->num_src_ports+i]);
		if ( i == chain->num_src_ports && (chain->flags & IP_FIREWALL_DST_RANGE) ) {
		    comma = ":";
		} else {
		    comma = ",";
		}
	    }
	    printf("\n");
	    chain = chain->next;
	}
    }
}

#ifdef LINUX
static void free_firewall_chain(struct ip_firewall **chainptr)
#else
static
void
free_firewall_chain(chainptr)
struct ip_firewall **chainptr;
#endif
{
    while ( *chainptr != NULL ) {
	struct ip_firewall *ftmp;
	ftmp = *chainptr;
	*chainptr = ftmp->next;
#ifdef LINUX
	kfree(ftmp);
#else
	free(ftmp,M_SOOPTS);
#endif
    }
}


#ifdef LINUX
static int add_to_chain(struct ip_firewall ***nextptr,
			struct ip_firewall *firewall)
#else
static
int
add_to_chain(nextptr,firewall)
struct ip_firewall ***nextptr;
struct ip_firewall *firewall;
#endif
{
    struct ip_firewall *ftmp;

#ifdef LINUX
    if (firewall->magic != IP_FIREWALL_MAGIC) {
      printf("add_to_chain: Got bad magic number in firewall, got %d, expecting %d\n",
	     firewall->magic,
	     IP_FIREWALL_MAGIC);
      return ( EINVAL );
    }
#endif LINUX

#ifdef LINUX
    ftmp = kmalloc( sizeof(struct ip_firewall), GFP_KERNEL);
#else
    ftmp = malloc( sizeof(struct ip_firewall), M_SOOPTS, M_DONTWAIT );
#endif
    if ( ftmp == NULL ) {
	printf("ip_firewall_ctl:  malloc said no\n");
	return( ENOSPC );
    }

#ifdef LINUX
    memcpy( ftmp, firewall, sizeof( struct ip_firewall ) );
#else
    bcopy( firewall, ftmp, sizeof( struct ip_firewall ) );
#endif
    ftmp->next = NULL;
    **nextptr = ftmp;
    *nextptr = &ftmp->next;

    return(0);
}

#ifdef LINUX
int ip_firewall_ctl(int stage, char *optval, int optlen)
#else
ip_firewall_ctl(stage,m)
int stage;		/* 0 -> blocking chain */
			/* 1 -> forwarding chain */
			/* 2 -> print chain on console */
			/* 3 -> flush entire chain */
struct mbuf *m;
#endif
{
    if ( stage == IP_PRINT_FIREWALLS ) {

	show_firewall_chain("blocking chain",ip_firewall_blocking_chain);
	show_firewall_chain("forwarding chain",ip_firewall_forwarding_chain);
	return(0);

    } else if ( stage == IP_FLUSH_FIREWALLS ) {

	free_firewall_chain(&ip_firewall_blocking_chain);
	free_firewall_chain(&ip_firewall_forwarding_chain);
	blocking_chain_next = &ip_firewall_blocking_chain;
	forwarding_chain_next = &ip_firewall_forwarding_chain;
	return(0);

    } else {

#ifdef LINUX
	if ( optval == 0 ) {
	    printf("ip_firewall_ctl:  NULL optval ptr\n");
	    return( EINVAL );
	}
#else 
	if ( m == 0 ) {
	    printf("ip_firewall_ctl:  NULL mbuf ptr\n");
	    return( EINVAL );
	}
#endif

	if ( stage == IP_CHECK_BLOCKING_FIREWALL || stage == IP_CHECK_FORWARDING_FIREWALL ) {

#ifdef LINUX
	    struct iphdr *ip;
	    if ( optlen < sizeof(struct iphdr) + 2 * sizeof(u_short) ) {
		printf("ip_firewall_ctl:  mbuf len=%d, want at least %d\n",optlen,sizeof(struct iphdr) + 2 * sizeof(u_short));
		return( EINVAL );
	    }
	    ip = (struct iphdr *) optval;
	    if ( ip->ihl != sizeof(struct iphdr) / sizeof(int) ) {
		printf("ip_firewall_ctl:  ip->ihl=%d, want %d\n",ip->ihl,sizeof(struct iphdr)/sizeof(int));
		return( EINVAL );
	    }
#else
	    struct ip *ip;
	    if ( m->m_len < sizeof(struct ip) + 2 * sizeof(u_short) ) {
		printf("ip_firewall_ctl:  mbuf len=%d, want at least %d\n",m->m_len,sizeof(struct ip) + 2 * sizeof(u_short));
		return( EINVAL );
	    }
	    ip = mtod(m,struct ip *);
	    if ( ip->ip_hl != sizeof(struct ip) / sizeof(int) ) {
		printf("ip_firewall_ctl:  ip->ip_hl=%d, want %d\n",ip->ip_hl,sizeof(struct ip)/sizeof(int));
		return( EINVAL );
	    }
#endif
	    if ( ip_firewall_check(ip,
		stage == IP_CHECK_BLOCKING_FIREWALL ? ip_firewall_blocking_chain : ip_firewall_forwarding_chain )
	    ) {
		return(0);
	    } else {
		return(EACCES);
	    }

	} else if ( stage == IP_ADD_BLOCKING_FIREWALL
	|| stage == IP_ADD_FORWARDING_FIREWALL ) {

	    struct ip_firewall *firewall;
#ifdef LINUX
	    if ( optlen != sizeof(struct ip_firewall) ) {
		printk("ip_firewall_ctl:  len=%d, want %d\n",optlen,sizeof(struct ip_firewall));
		return( EINVAL );
	    }

	    firewall = (struct ip_firewall *) optval;
#else
	    if ( m->m_len != sizeof(struct ip_firewall) ) {
		printf("ip_firewall_ctl:  len=%d, want %d\n",m->m_len,sizeof(struct ip_firewall));
		return( EINVAL );
	    }

	    firewall = mtod(m,struct ip_firewall*);
#endif
	    if ( (firewall->flags & ~IP_FIREWALL_FLAG_BITS) != 0 ) {
		printf("ip_firewall_ctl:  undefined flag bits set (flags=%x)\n",firewall->flags);
		return( EINVAL );
	    }

	    if ( (firewall->flags & IP_FIREWALL_SRC_RANGE) && firewall->num_src_ports < 2 ) {
		printf("ip_firewall_ctl:  SRC_RANGE set but num_src_ports=%d\n",firewall->num_src_ports);
		return( EINVAL );
	    }

	    if ( (firewall->flags & IP_FIREWALL_DST_RANGE) && firewall->num_dst_ports < 2 ) {
		printf("ip_firewall_ctl:  DST_RANGE set but num_dst_ports=%d\n",firewall->num_dst_ports);
		return( EINVAL );
	    }

	    if ( firewall->num_src_ports + firewall->num_dst_ports > IP_FIREWALL_MAX_PORTS ) {
		printf("ip_firewall_ctl:  too many ports (%d+%d)\n",firewall->num_src_ports,firewall->num_dst_ports);
		return( EINVAL );
	    }

#if 0
	    if ( (firewall->flags & IP_FIREWALL_KIND) == IP_FIREWALL_ICMP ) {
		printf("ip_firewall_ctl:  request for unsupported ICMP firewalling\n");
		return( EINVAL );
	    }

#endif
	    if ( stage == IP_ADD_BLOCKING_FIREWALL ) {
		return( add_to_chain(&blocking_chain_next,firewall) );
	    } else {
		return( add_to_chain(&forwarding_chain_next,firewall) );
	    }
	} else {
	    printf("ip_firewall_ctl:  unknown request %d\n",stage);
	    return(EINVAL);
	}
    }

    printf("ip_firewall_ctl:  stage = %d\n", stage);
    panic("ip_firewall_ctl:  how did we get here???");
}
