/*
 * Copyright (c) 1993 Daniel Boulet
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 *
 * Redistribution in binary form may occur without any restrictions.
 * Obviously, it would be nice if you gave credit where credit is due
 * but requiring it would be too onerous.
 *
 * This software is provided ``AS IS'' without any warranties of any kind.
 */

/*
 *  Linux port (c) 1994 Bob Beck
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 *
 * This software is provided ``AS IS'' without any warranties of any kind.
 *
 */

/* $Header: /usr/brule/dept/beck/ipfirewall/RCS/ip_firewall.h,v 1.1 1994/07/28 00:05:05 beck Exp $ */

/* $Log: ip_firewall.h,v $
 * Revision 1.1  1994/07/28  00:05:05  beck
 * Initial revision
 * */

/*
 * Format of an IP firewall descriptor
 *
 * src, dst, src_mask, dst_mask are always stored in network byte order.
 * flags and num_*_ports are stored in host byte order (of course).
 * Port numbers are stored in HOST byte order.
 */

struct ip_firewall {
#ifdef LINUX
    int magic;
#endif 
    struct ip_firewall *next;		/* Next firewall on chain */
#ifdef LINUX
    unsigned long src, dst;		/* Source and destination IP addr */
    unsigned long  src_mask, dst_mask;	/* Mask for src and dest IP addr */
#else
    struct in_addr src, dst;		/* Source and destination IP addr */
    struct in_addr src_mask, dst_mask;	/* Mask for src and dest IP addr */
#endif
    u_short flags;
#define IP_FIREWALL_UNIVERSAL	0	/* This is a universal packet firewall */
#define IP_FIREWALL_TCP		1	/* This is a TCP packet firewall */
#define IP_FIREWALL_UDP		2	/* This is a UDP packet firewall */
#define IP_FIREWALL_ICMP	3	/* This is a ICMP packet firewall */
#define IP_FIREWALL_KIND	3	/* Mask to isolate firewall kind */
#define IP_FIREWALL_ACCEPT	4	/* This is an accept firewall (as */
					/* opposed to a deny firewall) */
#define IP_FIREWALL_SRC_RANGE	8	/* The first two src ports are a min */
					/* and max range (stored in host byte */
					/* order). */
#define IP_FIREWALL_DST_RANGE	16	/* The first two dst ports are a min */
					/* and max range (stored in host byte */
					/* order). */
					/* firewall (ports[0] <= port <= ports[1]) */
#define IP_FIREWALL_FLAG_BITS	0x1f	/* All possible flag bits */
    u_short num_src_ports, num_dst_ports;/* # of src ports and # of dst ports */
    					/* in ports array (dst ports follow */
    					/* src ports; max of 10 ports in all; */
    					/* count of 0 means match all ports) */
#define IP_FIREWALL_MAX_PORTS	10	/* A reasonable maximum */
    u_short ports[IP_FIREWALL_MAX_PORTS]; /* Array of port numbers to match */
};

#ifdef LINUX
#define IP_FIREWALL_MAGIC 0x5303
#ifdef __KERNEL__
extern int ip_firewall_ctl(int stage, char *optval, int optlen);
extern int ip_firewall_check(struct iphdr *ip , struct ip_firewall *chain);
#endif
#endif

extern struct ip_firewall *ip_firewall_blocking_chain;
extern struct ip_firewall *ip_firewall_forwarding_chain;






